package istvan.xml;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.io.StringWriter;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

import javax.xml.transform.Source; 
import javax.xml.transform.stream.StreamSource; 
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer; 
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

/** An <tt>XSL</tt> filter utility to write <tt>XML</tt> 
  * <p>
  * Instances of this class can be used as simple filters to
  * transform <tt>XML</tt> source according to <tt>XSL</tt>
  * rules. This class is really a filter wrapper to a properly
  * configured <tt>javax.xml.transform.TransformerFactory</tt>.
  * You may also check the validity of the <tt>XML</tt> file
  * against a <tt>DTD</tt> definition.
  *
  * It is especially easy to use as an <i>un-deployed</i> or 
  * <i>on-the-fly</i> servlet filter:
  * <br><br>
  * <table border="3">
  * <tr>
  * <td bgcolor="ivory">
  * <font color="navy">
  * <pre>
  *
  *    ...
  *    String nl = System.getProperty("line.separator");
  *
  *    // get a new transformer
  *    XTransformer transformer = new XTransformer();
  *
  *    // get its <tt>XML</tt> writer
  *    Writer writer = transformer.getWriter();
  *
  *    // write your <tt>XML</tt> source
  *    writer.write("<b>&lt;?xml version=\"1.0\" encoding=\"ISO-8859-1\" ?&gt;</b>" + nl);
  *    // if you want to validate against a <tt>DTD</tt> ..
  *    writer.write("<b>&lt;!DOCTYPE <i>top-element</i> SYSTEM \"http://<i>the-server</i>/<i>location</i>/file.dtd\"&gt;</b>");
  *
  *    writer.write("<b>&lt;Book&gt;</b>" + nl);
  *    writer.write("<b>&lt;Title&gt;</b><i>War and Peace</i><b>&lt;/Title&gt;</b>" + nl);
  *    ...
  *
  *    // if you want to validate your <tt>XML</tt>
  *    XTransformer.ValidateHandler handler = transformer.<b>validate</b>();
  *    if (!handler.isClean()) {
  *       // then your <tt>XML</tt> has errors or warnings
  *       ...
  *    }
  *    // you could also validate against an <tt>XMLSchema</tt>
  *    ...
  *
  *    // get an input stream with the <tt>XSL</tt> rules
  *    // of course, it may come from a URL, socket, String ...
  *    // in <i>this</i> case it is a file ...
  *    FileInputStream xsl = new FileInputStream("<i>/usr/local/xsl/books.xsl</i>");   
  *
  *    // transform the source, the target is the console
  *    transformer.<b>transform</b>(System.out,xsl);
  *    xsl.close();
  *
  * </pre>
  * </font>
  * </td>
  * </tr>
  * </table>
  * <br><br>
  * If many streams are to be processed according to the same rules,
  * it is much more efficient to create a transformer object from
  * the rules and keep reusing it.
  * 
  * @author Istv&aacute;n T. Hern&aacute;dv&ouml;lgyi
  */
public class XTransformer {

   /** new line */
   private String nl = System.getProperty("line.separator");

   /** writer on the internal StringBuffer */
   private StringWriter writer = new StringWriter(65536);

   /** factory to provide transformers */
   private TransformerFactory tFactory = null;

   /** create a new transformer */
   public XTransformer() {
   }

   /** obtain the <tt>XML</tt> writer 
     *
     * @return <tt>XML</tt> writer
     */
   public Writer getWriter() {
      return writer;
   }

   /** reset the internal stream by removing all <tt>XML</tt> written so-far
     */
   public void reset() {
      writer.getBuffer().setLength(0);
   }

   /** validate the <tt>XML</tt> written so far
     * <p>
     *   The handler returned contains the <tt>DTD</tt> errors and
     *   warnings - <i>if any</i>.
     * </p>
     * @return handler passed to the <tt>SAX</tt> parser
     * @exception org.xml.sax.SAXException on errors occurring during parsing
     * @exception java.io.IOException on stream errors
     * @exception javax.xml.parsers.ParserConfigurationException on misconfiguration of the parser factory
     */
   public ValidateHandler validate() 
      throws SAXException, IOException, ParserConfigurationException  {

      SAXParserFactory spfact = SAXParserFactory.newInstance();

      spfact.setValidating(true);
      spfact.setNamespaceAware(true);

      SAXParser parser = spfact.newSAXParser();
      XMLReader reader = parser.getXMLReader();

      parser.parse(new InputSource(new StringReader(
         writer.getBuffer().toString())), handler);

      return handler;
   }

   /** transform the <tt>XML</tt> source written so far
     * <p>
     * This is simply a wrapper method and outputs the <tt>XML</tt>
     * source as it was written so far.
     * </p>
     * @param target output
     * @exception java.io.IOException on I/O error
     */
   public void transform(OutputStream target) throws IOException {
      target.write(writer.toString().getBytes());     
   }

   /** create a transformer object
     * <p>
     * This method creates a transformer object which can be used
     * with the corresponding transform methods. Creating a transformer
     * object and re-using it is much more efficient than reading
     * the transformation rules from a stream each time.
     * </p>
     * 
     * @param xsl <tt>XSL</tt> source
     * @return new transformer
     * @exception javax.xml.transform.TransformerConfigurationException if the transformer factory is not properly configured
     */
   public Transformer createTransformer(InputStream xsl)
      throws TransformerConfigurationException {
      if (tFactory == null) newFactory();

      Source xslSource = new StreamSource(xsl);
      
      return tFactory.newTransformer(xslSource);
   }

   /** transform the <tt>XML</tt> source written so far
     * <p>
     * This method transforms the <tt>XML</tt> source written so far
     * according to the (most likely <tt>XSL</tt>) rules of the transformer
     * </p>
     *
     * @param target output
     * @param xsl transformation rules
     * @exception java.io.IOException on I/O error
     * @exception javax.xml.transform.TransformerException if the transformation cannot succeed
     */
   public void transform(OutputStream target, Transformer xsl)
      throws TransformerException, IOException {
      xsl.transform(
          new StreamSource(new StringReader(writer.toString())),
          new StreamResult(target));
   }

   /** transform the <tt>XML</tt> source written so far
     * <p>
     * This method transforms the <tt>XML</tt> source written so far
     * according to the (most likely <tt>XSL</tt>) rules of the transformer
     * </p>
     *
     * @param target output
     * @param xsl transformation rules
     * @exception java.io.IOException on I/O error
     * @exception javax.xml.transform.TransformerException if the transformation cannot succeed
     */
   public void transform(Writer target, Transformer xsl)
      throws TransformerException, IOException {
      xsl.transform(
          new StreamSource(new StringReader(writer.toString())),
          new StreamResult(target));
   }

   /** transform the <tt>XML</tt> source written so far
     * <p>
     * This is simply a wrapper method and outputs the <tt>XML</tt>
     * source as it was written so far.
     * </p>
     * @param target output
     * @exception java.io.IOException on I/O error
     */
   public void transform(Writer target) throws IOException {
      target.write(writer.toString());     
   }

   /** create a new instance of the internal transformer factory
     * <p>
     * This method is only needed if the configuration has changed
     * or the internal transformer factory object is in an inconsistent 
     * state for some reason.
     * </p>
     *
     * @exception javax.xml.transform.TransformerConfigurationException if the transformer factory is not properly configured
     */
   public void newFactory() throws TransformerConfigurationException {
      try {
          tFactory = new org.apache.xalan.processor.TransformerFactoryImpl();
      } catch(Exception e) {
         tFactory = TransformerFactory.newInstance();
      }
   }

   /** transform the <tt>XML</tt> source written so far
     * <p>
     * This method transforms the <tt>XML</tt> source written so far
     * according to the (most likely <tt>XSL</tt>) rules read
     * from the stream <tt>xsl</tt>.
     * </p>
     *
     * @param target output
     * @param xsl transformation rules
     * @exception java.io.IOException on I/O error
     * @exception javax.xml.transform.TransformerConfigurationException if the transformer factory is not properly configured
     * @exception javax.xml.transform.TransformerException if the transformation cannot succeed
     */
   public void transform(OutputStream target, InputStream xsl) 
      throws IOException, TransformerConfigurationException,
             TransformerException {

      if (tFactory == null) newFactory();

      Source xslSource = new StreamSource(xsl);
      
      Transformer transformer = tFactory.newTransformer(xslSource);
      transformer.transform(
          new StreamSource(new StringReader(writer.toString())),
          new StreamResult(target));
   }
   
   /** transform the <tt>XML</tt> source written so far
     * <p>
     * This method transforms the <tt>XML</tt> source written so far
     * according to the (most likely <tt>XSL</tt>) rules read
     * from the stream <tt>xsl</tt>.
     * </p>
     *
     * @param target output
     * @param xsl transformation rules
     * @exception java.io.IOException on I/O error
     * @exception javax.xml.transform.TransformerConfigurationException if the transformer factory is not properly configured
     * @exception javax.xml.transform.TransformerException if the transformation cannot succeed
     */
   public void transform(Writer target, InputStream xsl) 
      throws IOException, TransformerConfigurationException,
             TransformerException {

      if (tFactory == null) newFactory();

      Source xslSource = new StreamSource(xsl);
      
      Transformer transformer = tFactory.newTransformer(xslSource);
      transformer.transform(
          new StreamSource(new StringReader(writer.toString())),
          new StreamResult(target));
   }


   /** command line transformer utility
     * <p>
     * This utility transforms an <tt>XML</tt> file according
     * to the rules of the given <tt>XSL</tt> file and dumps the
     * output on the console.
     * <pre>
     *  java  istvan.xml.XTransformer test.xml test.xsl [ -validate ]
     * </pre>
     *
     * </p>
     * @param argv[0] <tt>XML</tt> file name
     * @param argv[1] <tt>XSL</tt> file name
     * @param argv[2] the option <tt>-validate</tt> (<i>optional</i>)
     * @exception java.lang.Exception if anything goes wrong 
     */
   public static void main(String[] argv) throws Exception {

      if (argv.length < 2 || argv.length > 3) {
         System.out.println("java istvan.xml.XTransformer <xml> <xsl>" +
            " [ -validate ]");
         System.out.println("where <xml> and <xsl> are file names");
         return;
      }

      if (argv.length == 3) {
         if (!"-validate".equals(argv[2])) {
            System.out.println("?? "  + argv[2]);
            return;
         }
      }

      XTransformer transformer = new XTransformer(); 

      Writer writer = transformer.getWriter();

      FileInputStream xml = new FileInputStream(argv[0]);

      int c;
      while((c = xml.read()) != -1) {
         writer.write(c);
      }
      xml.close();

      if (argv.length==3) { 
         ValidateHandler handler = transformer.validate();

         if (!handler.isClean()) {
            if (handler.hasErrors()) System.out.println(handler.getErrors());
            if (handler.hasWarnings()) System.out.println(handler.getWarnings());
            return;
         }
      }

      FileInputStream xsl = new FileInputStream(argv[1]);

      Transformer xslt = transformer.createTransformer(xsl);
      xsl.close();

      transformer.transform(System.out,xslt);

   }

   /** validator for DTDs */
   private ValidateHandler handler = new ValidateHandler();

   /** A <tt>SAX</tt> <tt>DTD</tt> handler  
     * <p>
     * This handler can be used to obtain errors and warnings 
     * reported by the <tt>SAX</tt> parser.
     * </p>
     * @author Istv&aacute;n T. Hern&aacute;dv&ouml;lgyi
     * @see istvan.xml.XTransformer#validate
     */
   public class ValidateHandler extends DefaultHandler {

       /** no errors or warnings */
       private boolean clean = true;

       /** errors */
       private boolean errs = false;

       /** warnings */
       private boolean warrs = false;

       /** error buffer */
       private StringBuffer errors = new StringBuffer(16384);

       /** warning buffer */
       private StringBuffer warnings = new StringBuffer(16384);
       
       /** default constructor
         * <p>
         * It is <tt>private</tt> so cannot be created by other classes.
         * </p>
         */
       private ValidateHandler() {
       }

       /** clear error and warning flags */
       private void clearFlags() {
          clean = true;
          errors.setLength(0);
          warnings.setLength(0);
       }

       /** <tt>true</tt> if there are no errors or warnings
         * @return <tt>true</tt> if there are no errors or warnings
         */
       public boolean isClean() {
          return clean;
       }

       /** <tt>true</tt> if there are errors 
         * @return <tt>true</tt> if there are errors 
         */
       public boolean hasErrors() {
          return errs;
       }

       /** <tt>true</tt> if there are warnings 
         * @return <tt>true</tt> if there are warnings 
         */
       public boolean hasWarnings() {
          return warrs;
       }

       /** get errors
         * <p>
         * The errors are returned in <tt>XML</tt> format:
         * <pre>
         *   &lt;<b>errors</b>&gt;
         *      &lt;<b>error</b>&gt;<i>error 1</i>&lt;/<b>error</b>&gt;
         *      ...
         *   &lt;/<b>errors</b>&gt;
         * </pre>
         * where <tt><i>error 1</i></tt>, ... are errors reported by
         * the <tt>SAX</tt> parser.
         * </p>
         *
         * @return errors
         */ 
       public String getErrors() {
          return "<errors>" + nl + errors.toString() + "</errors>" + nl;
       }
      
       /** get warnings
         * <p>
         * The warnings are returned in <tt>XML</tt> format:
         * <pre>
         *   &lt;<b>warnings</b>&gt;
         *      &lt;<b>warning</b>&gt;<i>warning 1</i>&lt;/<b>warning</b>&gt;
         *      ...
         *   &lt;/<b>warnings</b>&gt;
         * </pre>
         * where <tt><i>warning 1</i></tt>, ... are errors reported by
         * the <tt>SAX</tt> parser.
         * </p>
         *
         * @return errors
         */ 
       public String getWarnings() {
          return "<warnings>" + nl + warnings.toString() + "</warnings>" + nl;
       }
      
       /** on error
         * <p>
         * The error callback method
         * </p>
         * @param exc <tt>SAX</tt> exception
         */
       public void error(SAXParseException exc) {
          errors.append("<error>" + exc.getMessage() + "</error>" + nl);
          clean = false;
          errs = true;
       }

       /** on warning
         * <p>
         * The warning callback method
         * </p>
         * @param exc <tt>SAX</tt> exception
         */
       public void warning(SAXParseException exc) {
          warnings.append("<warning>" + exc.getMessage() + "</warning>" + nl);
          clean = false;
          warrs = true;
       }
    }
}
